# noqa: D100

__title__ = "id-translation"
__description__ = "My personal little ML engineering library."
__version__ = "0.3.1"
__author__ = "Richard Sundqvist"
__copyright__ = "2022, Richard Sundqvist"
